(*
 * This file is part of Bolt.
 * Copyright (C) 2009-2011 Xavier Clerc.
 *
 * Bolt is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bolt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)


(* Definitions *)

type t = (string list) * (string list) * (Event.t -> string)

let layouts = Hashtbl.create 17

let register name layout =
  Hashtbl.replace layouts name layout

let register_unnamed layout =
  let i = ref 0 in
  while Hashtbl.mem layouts (string_of_int !i) do
    incr i
  done;
  let name = string_of_int !i in
  register name layout;
  name

let get name =
  Hashtbl.find layouts name


(* Predefined layouts *)

let decode s =
  try
    let idx = String.index s ':' in
    let len = String.length s in
    let key = String.sub s 0 idx in
    let size = String.sub s (succ idx) (len - idx - 1) in
    key, (try int_of_string size with _ -> 0)
  with Not_found -> s, 0

let assoc l s =
  try
    let key, size = decode s in
    let res = List.assoc key l in
    let left = size < 0 in
    let pad_size = (abs size) - (String.length res) in
    let pad_string = String.make (max 0 pad_size) ' ' in
    if left then pad_string ^ res else res ^ pad_string
  with _ ->
    ""

let render_bindings b fmt =
  let buffer = Buffer.create 256 in
  try
    Buffer.add_substitute buffer (assoc b) fmt;
    Buffer.contents buffer
  with _ -> fmt

let render fmt e =
  render_bindings (Event.bindings e) fmt
  
let simple =
  [], [], render "$(level:5) - $(message)"

let default =
  [], [], render "$(relative:-6) [$(file:-16) $(line:4)] $(level:5) - $(message)"

let pattern h f p =
  h, f, render p

type field_type =
  | Date
  | Int
  | Double
  | Hex
  | String

let string_of_field_type = function
  | Date -> "date"
  | Int -> "int"
  | Double -> "double"
  | Hex -> "hex"
  | String -> "string"

type event = {
    event_name : string;
    event_id : int;
    event_fields : (string * field_type) list;
  }

let event_def =
  let id = ref 0 in
  fun name l ->
    incr id;
    { event_name = name;
      event_id = !id;
      event_fields = l; }

let predefined_events = [
  event_def "PajeDefineContainerType"
    ["Alias", String; "ContainerType", String; "Name", String];
  event_def "PajeCreateContainer"
    ["Time", Date; "Alias", String; "Type", String; "Container", String; "Name", String];
  event_def "PajeDestroyContainer"
    ["Time", Date; "Name", String; "Type", String];
  event_def "PajeDefineEventType"
    ["Name", String; "ContainerType", String; "Alias", String];
  event_def "PajeDefineStateType"
    ["Name", String; "ContainerType", String; "Alias", String];
  event_def "PajeDefineVariableType"
    ["Name", String; "ContainerType", String; "Alias", String];
  event_def "PajeDefineLinkType"
    ["Name", String; "ContainerType", String; "SourceContainerType", String; "DestContainerType", String; "Alias", String];
  event_def "PajeDefineEntityValue"
    ["Name", String; "EntityType", String; "Alias", String];
  event_def "PajeNewEvent"
    ["Time", Date; "Type", String; "Container", String; "Value", String];
  event_def "PajeSetState"
    ["Time", Date; "Type", String; "Container", String; "Value", String];
  event_def "PajePushState"
    ["Time", Date; "Type", String; "Container", String; "Value", String];
  event_def "PajePopState"
    ["Time", Date; "Type", String; "Container", String];
  event_def "PajeSetVariable"
    ["Time", Date; "Type", String; "Container", String; "Value", Double];
  event_def "PajeAddVariable"
    ["Time", Date; "Type", String; "Container", String; "Value", Double];
  event_def "PajeSubVariable"
    ["Time", Date; "Type", String; "Container", String; "Value", Double];
  event_def "PajeStartLink"
    ["Time", Date; "Type", String; "Container", String; "SourceContainer", String; "Value", String; "Key", String];
  event_def "PajeEndLink"
    ["Time", Date; "Type", String; "Container", String; "DestContainer", String; "Value", String; "Key", String]
]

let string_of_event ev =
  let field_def (nam, typ) =
    Printf.sprintf "%%\t%s\t%s" nam (string_of_field_type typ) in
  ((Printf.sprintf "%%EventDef\t%s\t%d" ev.event_name ev.event_id)
   :: (List.map field_def ev.event_fields))
  @ ["%EndEventDef"]

let paje =
  (List.concat (List.map string_of_event predefined_events)),
  [],
  (fun e ->
    let default = function
      | Date -> string_of_float ((float e.Event.relative) /. 1000.)
      | Int -> "0"
      | Double -> "0.0"
      | Hex -> "00"
      | String -> "\"\"" in
    try
      let def =
        List.find
          (fun ed -> e.Event.message = ed.event_name)
          predefined_events in
      let fields =
        List.map
          (fun (nam, typ) ->
            try
              let res = List.assoc nam e.Event.properties in
              if typ = String then Printf.sprintf "%S" res else res
            with Not_found -> default typ)
          def.event_fields in
      let fields = (string_of_int def.event_id) :: fields in
      String.concat " " fields
    with _ -> "")

let style = [
  "<style type=\"text/css\">" ;
  "table.sample {" ;
  "  border-width: 1px;" ;
  "  border-style: solid;" ;
  "  border-color: black;" ;
  "  border-collapse: collapse;" ;
  "  background-color: white;" ;
  "}" ;
  "table.sample th {" ;
  "  border-width: 1px;" ;
  "  padding: 3px;" ;
  "  border-style: solid;" ;
  "  border-color: black;" ;
  "  background-color: white;" ;
  "}" ;
  "table.sample td {" ;
  "  border-width: 1px;" ;
  "  padding: 3px;" ;
  "  border-style: solid;" ;
  "  border-color: black;" ;
  "  background-color: white;" ;
  "}" ;
  "</style>"
]

let html =
  [ "<html>" ;
    "<head>" ;
    "<title>Log</title>" ] @
  style @
  [ "</head>" ;
    "<body>" ;
    "<p align=\"center\">" ;
    "<table class=\"sample\">" ;
    ("<tr>" ^
     "<th>Identifier</th>" ^
     "<th>Time</th>" ^
     "<th>Millis</th>" ^
     "<th>Level</th>" ^
     "<th>File</th>" ^
     "<th>Message</th>" ^
     "</tr>") ],
  [ "</table>" ;
    "</p>" ;
    ("<p style=\"font-size: smaller; text-align: center;\">Generated by <a href=\"http://bolt.x9c.fr\">Bolt " ^ Version.value ^ "</a></p>") ;
    "</body>" ;
    "</html>" ],
  render ("<tr style=\"font-family: monospace;\">" ^
          "<td align=\"right\">$(id)</td>" ^
          "<td>$(hour):$(min):$(sec)</td>" ^
          "<td align=\"right\">$(relative)</td>" ^
          "<td>$(level)</td>" ^
          "<td><pre>$(filebase:-24) $(line:4)</pre></td>" ^
          "<td>$(message)</td>" ^
          "</tr>")

let protect_cdata s =
  let len = String.length s in
  let buf = Buffer.create len in
  let i = ref 0 in
  while (!i < len) do
    if (!i >= 2) && (s.[!i - 2] = ']') && (s.[!i - 1] = ']') && (s.[!i] = '>') then
      Buffer.add_string buf " >"
    else
      Buffer.add_char buf s.[!i];
    incr i
  done;
  Buffer.contents buf

let xml =
  [],
  [],
  (fun e ->
    let properties =
      List.map
        (fun (k, v) -> Printf.sprintf "  <log4j:data name=\"%s\" value=\"%s\"/>\n" k v)
        e.Event.properties in
    render_bindings
      (List.map
         (fun (k, v) -> (k, (if k = "message" then protect_cdata v else v)))
         (Event.bindings e))
      ("<log4j:event logger=\"$(logger)\" level=\"$(level)\" thread=\"$(thread)\" timestamp=\"$(time)\">\n" ^
       "<log4j:message><![CDATA[$(message)]]></log4j:message>\n" ^
       (match e.Event.error with
       | Some (_, s) -> "<log4j:throwable><![CDATA[" ^ (protect_cdata s) ^ "]]></log4j:throwable>\n"
       | None -> "") ^
       "<log4j:locationInfo class=\"Unknown\" method=\"unknown()\" file=\"$(file)\" line=\"$(line)\"/>\n" ^
       "<log4j:properties>\n" ^
       (String.concat "" properties) ^
       "</log4j:properties>\n" ^
       "</log4j:event>\n"))

let csv sep l =
  let render e =
    let b = Event.bindings e in
    let values = List.map (fun k -> try List.assoc k b with _ -> "") l in
    String.concat sep values in
  ([], [], render)

let () =
  List.iter
    (fun (x, y) -> register x y)
    [ "simple",  simple ;
      "default", default ;
      "html",    html ;
      "paje",    paje ;
      "xml",     xml ]
